# -*- coding: utf-8 -*-
from iktomi.cms.loner import Loner
from iktomi.cms.publishing.loner import PublishLoner, PublishLonerNoState
from .i18n_stream import I18nStreamMixin


class I18nLonerMixin(I18nStreamMixin):

    def url_for(self, env, name=None, **kwargs):
        kwargs.setdefault('version', getattr(env, 'version', None) or self.versions[0][0])
        kwargs.setdefault('lang', getattr(env, 'lang', None) or self.langs[0][0])
        return Loner.url_for(self, env, name, **kwargs)


class I18nPublishLonerNoState(I18nLonerMixin, PublishLonerNoState):
    pass


class I18nPublishLoner(I18nLonerMixin, PublishLoner):
    pass





# -*- coding: utf-8 -*-
import six

from iktomi.forms import Field, FieldSet, FieldList
from iktomi.cms.forms import convs
from iktomi.cms.list_edit import ListItemForm as BaseListItemForm, \
        ListEditAction as BaseListEditAction, ListItemModelChoice

from .item_lock import ModelLockedByOther, ModelLockError
from .flashmessages import flash
from webob.exc import HTTPForbidden


class ListItemForm(BaseListItemForm):

    ordering_field = 'order'
    template = 'stream_sortables.html'

    def __init__(self, *args, **kwargs):
        self.ordering_field = kwargs.pop('ordering_field', self.ordering_field)
        BaseListItemForm.__init__(self, *args, **kwargs)

    fields = [
            FieldList(
                'items',
                widget=FieldList.widget(
                    template='list_item_form'),
                field=FieldSet(
                    'item',
                    fields=[
                        Field('item',
                              conv=ListItemModelChoice(required=False),
                              label=u'item'),
                        Field("order",
                              conv=convs.Int(),
                              label=u'Порядок'),
                    ]
                ),
            ),
        ]

    def initial_for_item(self, item):
        return {"item": item,
                "order": getattr(item, self.ordering_field)}


class ListEditAction(BaseListEditAction):

    def set_order(self, item, ordering_field, position):
        setattr(item, ordering_field, position)

    def list_edit(self, env, data):
        if not self.save_allowed(env):
            raise HTTPForbidden()

        list_item_form = self.ListItemForm(env)
        ordering_field = list_item_form.ordering_field
        if list_item_form is not None:

            if list_item_form.accept(env.request.POST):
                modified_items = []
                lock_messages = []
                edit_sessions = {}
                for value in list_item_form.python_data['items']:
                    item = value['item']
                    if getattr(item, ordering_field) != value["order"]:
                        try:
                            edit_sessions[item] = env.item_lock.create(item)
                        except ModelLockedByOther as e:
                            lock_messages.append(six.u(e))
                        except ModelLockError as e:
                            lock_messages.append(six.u(e))
                        self.set_order(item, ordering_field, value['order'])
                        modified_items.append(item)

                if len(lock_messages) > 0:
                    env.db.rollback()
                    flash(env, u'Изменения не были сохранены:\n%s' % 
                               '\n'.join(lock_messages),
                        'failure')
                elif len(modified_items) > 0:
                    env.db.commit()
                    flash(env,
                          '\n'.join([u'Объект (%s) сохранен' % (item,)
                                     for item in modified_items]),
                          'success')
                    [env.item_lock.remove(item, data.edit_session)
                     for item, data.edit_session in edit_sessions.items()]
                return env.json({'success': True })
            else:
                flash(env,
                    u'Изменения не были сохранены из-за ошибок в форме:\n%s' %
                    '\n'.join(list_item_form.errors), 'failure')

        return env.json({'success': False})
    __call__ = list_edit




# -*- coding: utf-8 -*-
from iktomi.cms.forms.fields import Field
from iktomi.cms.forms import widgets

__all__ = ['EditorNoteField']


class EditorNoteField(Field):

    permissions = 'r'
    widget = widgets.Widget(template="widgets/editor_notes")

    def get_data(self):
        item = self.form.item
        if item is None or item.id is None:
            return []
        EditorNote = self.model
        env = self.env
        return EditorNote.get_for_item(env.db, env.stream.uid(env, version=False), item)

    @property
    def submit_url(self):
        return self.env.url_for('post_note')



# -*- coding: utf-8 -*-
from functools import partial
import json

from jinja2 import Markup
from webob.exc import HTTPSeeOther

from iktomi import web
from iktomi.utils.storage import storage_cached_property, storage_method, storage_property
from iktomi.templates import BoundTemplate
from iktomi.utils import cached_property, quote_js
# XXX include i18n package into iktomi
#from i18n.translation import get_translations


class AdminEnvironment(web.AppEnvironment):

    # Provide following attributes in your subclass:
    #
    # cfg = cfg
    # redis = redis_client
    #   or
    # cache = memcached_client
    # file_manager = file_manager
    # streams = streams
    # models = models
    # static = static
    # db_maker = db_maker
    # template_loader = template_loader
    # get_help = get_help
    #
    # @storage_cached_property
    # def template(storage):
    #     return BoundTemplate(storage, storage.template_loader)

    login = None
    lang = None
    site_lang = 'ru'

    @cached_property
    def url_for(self):
        return self.root.build_url

    @cached_property
    def url_for_static(self):
        return self.static.url_for_static

    @storage_property
    def render_to_string(storage):
        return storage.template.render

    @storage_property
    def render_to_response(storage):
        return storage.template.render_to_response

    @storage_method
    def redirect_to(storage, name, qs, **kwargs):
        url = storage.url_for(name, **kwargs)
        if qs:
            url = url.qs_set(qs)
        return HTTPSeeOther(location=str(url))

    @storage_method
    def json(self, data):
        if getattr(self, '_flash', None):
            assert '_flashmessages' not in data
            data = dict(data, _flashmessages=self._flash)
        return web.Response(json.dumps(data), content_type="application/json")

    @cached_property
    def db(self):
        return self.db_maker()

    @storage_cached_property
    def auth_model(storage):
        return storage.models_.admin.AdminUser

    @storage_cached_property
    def draft_form_model(storage):
        return storage.models_.admin.DraftForm

    @storage_cached_property
    def edit_log_model(storage):
        return storage.models_.admin.EditLog

    @storage_cached_property
    def object_tray_model(storage):
        return storage.models_.admin.ObjectTray

    @storage_cached_property
    def tray_model(storage):
        return storage.models_.admin.Tray

    @storage_method
    def get_edit_url(storage, x):
        return storage.streams.get_edit_url(storage, x)

    def __init__(self, *args, **kwargs):
        web.AppEnvironment.__init__(self, *args, **kwargs)
        self.models_ = self.models # XXX bad naming!!

    # XXX include i18n package into iktomi
    #def get_translations(self, lang):
    #    return get_translations(self.cfg.I18N_TRANSLATIONS_DIR, lang,
    #                            ['iktomi-forms'])

    #@cached_property
    #def _translations(self):
    #    return self.get_translations(self.site_lang)

    #@storage_method
    #def gettext(self, msgid):
    #    message = self._translations.gettext(unicode(msgid))
    #    if isinstance(msgid, Markup):
    #        message = Markup(message)
    #    return message

    #@storage_method
    #def ngettext(self, msgid1, msgid2, n):
    #    message = self._translations.ngettext(unicode(msgid1),
    #                                          unicode(msgid2), n)
    #    if isinstance(msgid1, Markup):
    #        message = Markup(message)
    #    return message

    # XXX include RedisItemLock into iktomi
    #@storage_cached_property
    #def item_lock(storage):
    #    return RedisItemLock(storage)


class AdminContext(object):

    def __init__(self, env):
        self.env = env

    @cached_property
    def top_menu(self):
        # redefine this in your app
        raise NotImplementedError

    def item_trays(self, stream, item):
        ObjectTray = self.env.object_tray_model
        return self.env.db.query(ObjectTray)\
                          .filter_by(stream_name=stream.uid(self.env, version=False),
                                     object_id=item.id).all()

    @cached_property
    def users(self):
        AdminUser = self.env.auth_model
        return self.env.db.query(AdminUser).filter_by(active=True).all()

    @cached_property
    def user_tray(self):
        Tray = self.env.tray_model
        user = self.env.user
        tray = self.env.db.query(Tray).filter_by(editor=user).first()
        if tray is None:
            tray = Tray(editor=user, title=user.name or user.login)
            self.env.db.add(tray)
            self.env.db.commit()
        return tray

    @property
    def flashmessages(self):
        return getattr(self.env, '_flash', [])


class AdminBoundTemplate(BoundTemplate):

    # packer = packer
    # Context = Context
    constant_template_vars = {
        'quote_js': quote_js,
    }

    def get_template_vars(self):
        d = dict(
            self.constant_template_vars,
            env=self.env,
            user=getattr(self.env, 'user', None),
            url_for=self.env.url_for,
            url_for_static=self.env.url_for_static,
            packed_js_tag=partial(self.packer.js_tag, self.env),
            packed_css_tag=partial(self.packer.css_tag, self.env),
            context=self.Context(self.env),
        )
        import_settings = getattr(self.env.cfg, 'TEMPLATE_IMPORT_SETTINGS', [])
        for key in import_settings:
            d[key] = getattr(self.env.cfg, key)
        return d

    def render(self, template_name, __data=None, **kw):
        r = BoundTemplate.render(self, template_name, __data, **kw)
        return Markup(r)



# -*- coding: utf-8 -*-
from datetime import datetime

from sqlalchemy import Column, ForeignKey, Integer, DateTime, \
        Index, String, Text
from sqlalchemy.orm import relationship, object_session
from sqlalchemy import func

from iktomi.utils import cached_property
from iktomi.cms.models.base import register_model

__all__ = ['Tray']


@register_model('BaseModel')
def ObjectTray(models):
    id = Column(Integer, primary_key=True)
    stream_name = Column(String(50), nullable=False, default='')
    # object id can be string, so we use string here
    object_id = Column(String(100), nullable=True)

    tray_id = Column(Integer, ForeignKey('Tray.id'))
    tray = relationship('Tray')
    sender_id = Column(Integer, ForeignKey(models.AdminUser.id),
                       nullable=True)
    sender = relationship(models.AdminUser)

    comment = Column(Text, nullable=False, default='')
    created_dt = Column(DateTime, default=datetime.now, nullable=False)

    Index('tray_index', stream_name, object_id)

    def can_delete(self, user):
        if self.tray.editor is not None:
            return user in (self.tray.editor, self.sender)
        return True


@register_model('BaseModel')
def Tray(models):

    id = Column(Integer, primary_key=True)
    title = Column(String(250), nullable=False)
    editor_id = Column(Integer, ForeignKey(models.AdminUser.id),
                       nullable=True, unique=True)
    editor = relationship(models.AdminUser)

    @cached_property
    def object_count(self):
        session = object_session(self)
        ObjectTray = self.models.ObjectTray
        return session.query(func.count(ObjectTray.id))\
                      .filter(ObjectTray.tray==self)\
                      .scalar()




# -*- coding: utf-8 -*-

from iktomi.utils import cached_property
from iktomi.cms.stream_sortables import ListEditAction, ListItemForm


class PublishSortForm(ListItemForm):

    template = 'publish_stream_sortables.html'


class PublishSortAction(ListEditAction):

    def save_allowed(self, env):
        return self.stream.has_permission(env, 'w') and \
                self.stream.has_permission(env, 'p')

    @cached_property
    def ListItemForm(self):
        return getattr(self.stream.config, 'ListItemForm', PublishSortForm)

    def set_order(self, item, ordering_field, position):
        setattr(item, ordering_field, position)
        setattr(item._item_version('front'), ordering_field, position)




# -*- coding: utf-8 -*-
import six

from iktomi.forms.fields import *
from iktomi.forms.fields import __all__ as _all1
from iktomi.forms.form import Form

from datetime import datetime, timedelta
from sqlalchemy import desc
from sqlalchemy.orm.util import identity_key
from sqlalchemy.orm.exc import UnmappedInstanceError

from iktomi.cms.forms import convs, widgets
from iktomi.cms.publishing.model import WithState
from iktomi.cms.edit_log.models import _get_field_data, make_diff
from iktomi.forms import fields
_all2 = locals().keys()

# place this after _all2 to add them to __all__
from .files import AjaxImageField, AjaxFileField

def optional_unicode(value):
    return six.u(value) if hasattr(value, '__unicode__') else ''

class DiffFieldSetMixIn(object):

    def get_diff(form1, form2):
        # diff for edit log
        fields = sum([x.field_names for x in form1.fields], [])
        diffs = []
        for field_name in fields:
            if not field_name:
                continue
            field1 = form1.get_field(field_name)
            field2 = form2.get_field(field_name)
            if field2.widget.render_type == "hidden":
                continue
            if hasattr(field1, 'get_diff'):
                child_diff = field1.get_diff(field2)
                if child_diff is not None:
                    diffs.append(child_diff)
            elif hasattr(field1.widget, 'get_diff'):
                child_diff = field1.widget.get_diff(field1, field2)
                if child_diff is not None:
                    diffs.append(child_diff)
            else:
                data1 = _get_field_data(form1, field1)
                data2 = _get_field_data(form2, field2)
                if data1 != data2:
                    diff = make_diff(field1, field2,
                                     changed=True)
                    diffs.append(diff)
        if diffs:
            if isinstance(form1, Form):
                label = name = ''
                before = lambda: optional_unicode(form1.item)
                after = lambda: optional_unicode(form2.item)
            else:
                label = form1.label or form1.name
                name = form1.input_name
                before = lambda: optional_unicode(form1.clean_value)
                after = lambda: optional_unicode(form2.clean_value)
            return dict(label=label or '',
                        name=name or '',
                        before=before,
                        after=after,
                        children=diffs,
                        changed=True)


class FieldSet(FieldSet, DiffFieldSetMixIn):
    pass


class FieldBlock(FieldBlock, DiffFieldSetMixIn):
    pass


class FieldList(FieldList):

    @staticmethod
    def _is_pair(field1, field2):
        if field1.conv.__class__ == field2.conv.__class__:
            if isinstance(field1.conv, convs.ModelDictConv):
                try:
                    ident1 = identity_key(instance=field1.clean_value)[1]
                    ident2 = identity_key(instance=field2.clean_value)[1]
                    return ident1 == ident2
                except UnmappedInstanceError:
                    pass
            # XXX how to implement indication in this case?
            return True
        return False

    def get_diff(fieldlist1, fieldlist2):
        reordered = False
        fields1 = []
        fields2 = []
        for index in fieldlist1.form.raw_data.getall(fieldlist1.indices_input_name):
            fields1.append(fieldlist1.field(name=index))
        for index in fieldlist2.form.raw_data.getall(fieldlist2.indices_input_name):
            fields2.append(fieldlist2.field(name=index))

        diffs = []
        while fields1:
            field2 = None
            field1 = fields1.pop(0)
            if fields2 and fieldlist1._is_pair(field1, fields2[0]):
                field2 = fields2.pop(0)

            if field1.widget.render_type == "hidden":
                continue
            if field2 is not None:
                if hasattr(field1, 'get_diff'):
                    diff = field1.get_diff(field2)
                    if diff is not None:
                        diffs.append(diff)
                elif hasattr(field1.widget, 'get_diff'):
                    diff = field1.widget.get_diff(field1, field2)
                    if diff is not None:
                        diffs.append(diff)
                else:
                    data1 = _get_field_data(fieldlist1, field1)
                    data2 = _get_field_data(fieldlist2, field2)
                    if data1 != data2:
                        diff = make_diff(field1, field2,
                                        changed=True)
                        diffs.append(diff)
            else:
                diff = make_diff(field1, None,
                                 changed=False)
                diffs.append(diff)
                reordered = True

        while fields2:
            field2 = fields2.pop(0)
            diff = make_diff(None, field2,
                            changed=True)
            diffs.append(diff)

        if not reordered:
            diffs = [x for x in diffs if x['changed'] or x]
        if diffs:
            return dict(label=fieldlist1.label or '',
                        name=fieldlist1.input_name,
                        before=lambda: '',
                        after=lambda: '',
                        children=diffs,
                        changed=True)


def SplitDateTimeField(name, label, required=True,
                       get_initial=datetime.now,
                       template='widgets/fieldset-line'):
    return FieldSet(
        name,
        widget=widgets.FieldSetWidget(js_block='datetime',
                                      template=template),
        conv=convs.SplitDateTime(required=required),
        fields=[Field('date',
                      conv=convs.Date(required=required),
                      widget=widgets.Calendar(js_block='calendar-simple')),
                Field('time',
                      conv=convs.Time(required=required),
                      widget=widgets.TextInput(classname='timeinput'))],
        get_initial=get_initial,
        label=label)


class FieldBlock(fields.FieldBlock):

    widget = widgets.FieldBlockWidget


class DateFromTo(FieldSet):

    conv=convs.DateIntervalConv
    label=u'Дата'

    fields = [
        Field('since',
              conv=convs.Date(required=False),
              widget=widgets.Calendar,
              label=u'с'),
        Field('till',
              conv=convs.Date(required=False),
              widget=widgets.Calendar,
              label=u'по'),
    ]

    def filter_query(self, query, field, value):
        model = self.form.model
        if value['since'] is not None:
            query = query.filter(getattr(model, self.name)>=value['since'])
        if value['till'] is not None:
            next_day = value['till'] + timedelta(days=1)
            query = query.filter(getattr(model, self.name)<next_day)
        return query


_state_choices = ((WithState.PRIVATE, u'Скрытое'),
                  (WithState.PUBLIC, u'Опубликованное'))

def StateSelectField():
    return Field('state',
              conv=convs.EnumChoice(choices=_state_choices,
                                    conv=convs.Int()),
              widget=widgets.LabelSelect(render_type='full-width',
                                         null_label=u'Все'),
              )

def IdField(name='id', conv=convs.Int):
    return Field(name,
                 conv=convs.Int(required=False),
                 widget=widgets.TextInput(template="widgets/id_field",
                                          classname="small"),
                 label=u'Идентификатор',
                 )


class SortConverter(convs.EnumChoice):

    def to_python(self, value):
        value = convs.EnumChoice.to_python(self, value)
        value = value or self.field.get_initial()
        self.field.set_raw_value(self.field.form.raw_data,
                                 self.from_python(value))
        return value


class SortWidget(widgets.Select):

    classname='js-sort-field'
    render_type="hidden"

    def get_options(self, value):
        options = widgets.Select.get_options(self, value)
        if not value:
            initial_value = self.field.conv.from_python(
                    self.field.get_initial())
            for option in options:
                option['selected'] = option['value'] == initial_value
        return options


class SortField(Field):

    conv = SortConverter()
    widget = SortWidget()
    # (db column, list_field name)
    choices = (('id', 'id'),)

    def __init__(self, *args, **kwargs):
        Field.__init__(self, *args, **kwargs)
        choices = sum([[(k, v), ('-'+k, '-'+v)]
                       for k, v in self.choices], [])
        self.conv = self.conv(choices=choices, required=True)

    def filter_query(self, query, field, value):
        is_desc = value.startswith('-')
        value = value.lstrip('-')
        method = getattr(self.form, 'order_by__' + value,
                         getattr(self, 'order_by__' + value,
                             self.order_by_default))
        return method(query, value, is_desc)

    def order_by_default(self, query, value, is_desc):
        value = getattr(self.form.model, value)
        if is_desc:
            value = desc(value)
        return query.order_by(value)

    def set_raw_value(self, raw_data, value):
        if value == self.conv.from_python(self.get_initial()):
            raw_data[self.input_name] = ''
        else:
            raw_data[self.input_name] = value


object_ref_fields = [
    Field('stream_name',
          conv=convs.Char(convs.length(0, 100), required=True)),
    Field('object_id',
          conv=convs.Char(convs.length(0, 100), required=True)),
]


# Expose all variables defined after imports and all variables imported from
# parent module
__all__ = [x for x
           in set(locals().keys()) - (set(_all2) - set(_all1))
           if not x.startswith('_')]
del _all1, _all2



# -*- coding: utf-8 -*-
from iktomi import web
from iktomi.cms.stream_handlers import PrepareItemHandler, EditItemHandler,\
        DeleteItemHandler, StreamListHandler
from iktomi.cms.stream import Stream, ListField, FilterForm
from iktomi.cms.stream_actions import PostAction
from iktomi.cms.flashmessages import flash
from iktomi.cms.publishing.model import _AdminReplicated
from iktomi.utils import cached_property
from iktomi.utils.storage import VersionedStorage
from jinja2 import Markup


class PublishStreamListHandler(StreamListHandler):

    def prepare_data(self, env, data):
        data = StreamListHandler.prepare_data(self, env, data)
        data['allow_add'] = data['allow_add'] and env.version == 'admin'
        return data

    def list_form_data(self, env, paginator, filter_data):
        if env.version == 'admin':
            return StreamListHandler.list_form_data(
                    self, env, paginator, filter_data)
        return {}


class PublishItemHandler(EditItemHandler):

    def save_allowed(self, env, item=None):
        return env.version == 'admin' and \
            EditItemHandler.save_allowed(self, env, item)

    def process_item_template_data(self, env, td):
        # Filter buttons
        td['version'] = env.version
        changes = self.changed_fields(env, td['item'], td['form'])
        td['form'].changed_fields = changes
        return td

    def create_allowed(self, env):
        return self.stream.has_permission(env, 'c') and env.version == 'admin'

    def delete_allowed(self, env, item=None):
        return self.stream.has_permission(env, 'd') and env.version == 'admin'

    def get_front_item_form(self, env, item):
        front_models = _AdminReplicated.front
        if getattr(env, "lang", None):
            front_models = getattr(front_models, env.lang)

        front_env = VersionedStorage(db=env.db,
                                     models=front_models,
                                     user=env.user,
                                     version='front')
        front_env._storage._parent_storage = env

        form_cls = self.stream.config.ItemForm(front_env, {}, item._front_item)
        form = form_cls.load_initial(env, item._front_item, initial={}, permissions='r')
        form.model = self.stream.get_model(env)
        return form

    def _collect_changed_fields(self, diff):
        names = sum([self._collect_changed_fields(subdiff)
                     for subdiff in diff.get('children', [])], [])
        if names:
            return names
        if diff['name'] and diff['changed']:
            return [diff['name']]
        return []

    def changed_fields(self, env, item, admin_form):
        if env.version == 'front' or \
           not item.has_unpublished_changes or \
           not admin_form.is_valid or \
           (hasattr(item, 'state') and \
            item.state not in (item.PUBLIC, item.PRIVATE)):
                #(not item.has_unpublished_changes and \
                # getattr(admin_form, 'draft', None) is None):
            return []

        front_form = self.get_front_item_form(env, item)
        diff = front_form.get_diff(admin_form)
        if diff is None:
            return []
        return self._collect_changed_fields(diff)

    def edit_item_handler(self, env, data):
        # XXX hack!
        data._has_unpublished_changes = False
        if data.item is not None:
            data._has_unpublished_changes = data.item.has_unpublished_changes
        return EditItemHandler.edit_item_handler(self, env, data)
    __call__ = edit_item_handler

    def save_log_item(self, env, data, log, item):
        EditItemHandler.save_log_item(self, env, data, log, item)
        if not log.data_changed:
            # if there were no changes, revert has_unpublished_changes to
            # initial state
            item.has_unpublished_changes = data._has_unpublished_changes
            env.db.commit()


class PublishAction(PostAction):

    action = 'publish'
    cls = 'publish'
    display = True
    for_item = True
    allowed_for_new = False
    accepts_item_form = False
    title = u'Опубликовать'
    hint = u'Перенести изменения из редакторской версии на опубликованную ' \
           u'и сделать её доступной для просмотра на сайте'
    PrepareItemHandler = PrepareItemHandler

    @property
    def app(self):
        return web.match('/<idconv:item>/%s' % self.action, self.action,
                         convs={'idconv': self.stream.id_converter}) | \
               web.method('POST', strict=True) | \
               self.PrepareItemHandler(self) | self

    def admin_publish(self, env, data):
        self.stream.insure_has_permission(env, 'p')
        if data.lock_message:
            self.stream.rollback_due_lock_lost(env, data.item)
            return env.json({'success': False,
                             'error': 'item_lock',
                             'lock_message': data.lock_message})

        log = self.stream.create_log_entry(env, data.item, 'publish')
        if log is not None:
            env.db.add(log)

        data.item.publish()
        flash(env, u'Объект «%s» опубликован' % data.item, 'success')
        env.db.commit()

        url = self.stream.url_for(env, 'item', item=data.item.id)
        return env.json({'success': True,
                         'item_id':data.item.id,
                         'location': url})
    __call__ = admin_publish

    def is_available(self, env, item):
        return item.id is not None and \
            (not hasattr(item, 'state') or item.existing) and \
            (item.has_unpublished_changes or \
                (hasattr(item, 'state') and not item.public)) and \
            self.stream.has_permission(env, 'p') and \
            env.version == 'admin'


class UnpublishAction(PostAction):

    allowed_for_new = False
    accepts_item_form = False
    action = 'unpublish'
    cls = 'unpublish'
    title = Markup(u'Снять<br/> с публикации')
    PrepareItemHandler = PrepareItemHandler

    @property
    def app(self):
        return web.match('/<idconv:item>/%s' % self.action, self.action,
                         convs={'idconv': self.stream.id_converter}) | \
               web.method('POST', strict=True) | \
               self.PrepareItemHandler(self) | self

    def unpublish(self, env, data):
        self.stream.insure_has_permission(env, 'p')
        if data.lock_message:
            self.stream.rollback_due_lock_lost(env, data.item)
            return env.json({'success': False,
                             'error': 'item_lock',
                             'lock_message': data.lock_message})

        log = self.stream.create_log_entry(env, data.item, 'unpublish')
        if log is not None:
            env.db.add(log)

        data.item.unpublish()
        flash(env, u'Объект «%s» снят с публикации' % data.item, 'success')
        env.db.commit()

        url = self.stream.url_for(env, 'item', item=data.item.id)
        return env.json({'success': True,
                         'item_id':data.item.id,
                         'location': url})
    __call__ = unpublish

    def is_available(self, env, item):
        return hasattr(item, 'state') and \
                item.state == item.PUBLIC and \
                self.stream.has_permission(env, 'p') and \
                env.version == 'admin'


class RevertAction(PostAction):

    allowed_for_new = False
    accepts_item_form = False
    action = 'revert'
    cls = 'revert'
    title = Markup(u'Восстановить<br/> из опубликованной')
    hint = u'Отменить изменения, сделанные после публикации'
    PrepareItemHandler = PrepareItemHandler

    @property
    def app(self):
        return web.match('/<idconv:item>/%s' % self.action, self.action,
                         convs={'idconv': self.stream.id_converter}) | \
               web.method('POST', strict=True) | \
               self.PrepareItemHandler(self) | self

    def _clean_item_data(self, stream, env, item):
        form_cls = stream.config.ItemForm
        form = form_cls.load_initial(env, item, initial={}, permissions='r')
        return form.raw_data.items()

    def revert(self, env, data):
        self.stream.insure_has_permission(env, 'p')
        if data.lock_message:
            self.stream.rollback_due_lock_lost(env, data.item)
            return env.json({'success': False,
                             'error': 'item_lock',
                             'lock_message': data.lock_message})


        log = self.stream.create_log_entry(env, data.item, 'revert')
        if log is not None:
            log.before = self._clean_item_data(self.stream, env, data.item)
            env.db.add(log)

        data.item.revert_to_published()

        DraftForm = getattr(env, 'draft_form_model', None)
        if DraftForm is not None:
            draft = DraftForm.get_for_item(env.db,
                                           self.stream.uid(env),
                                           data.item, env.user)
            if draft is not None:
                env.db.delete(draft)

        flash(env, u'Объект «%s» восстановлен из опубликованной версии'
                    % data.item, 'success')
        env.db.commit()

        if log is not None:
            log.after = self._clean_item_data(self.stream, env, data.item)
            env.db.commit()


        url = self.stream.url_for(env, 'item', item=data.item.id)
        return env.json({'success': True,
                         'location': url})
    __call__ = revert

    def is_available(self, env, item):
        if item.id is None or (
                env.version == 'admin' and \
                hasattr(item._front_item, 'state') and \
                item._front_item.state not in (item.PUBLIC, item.PRIVATE)
                ) or not (self.stream.has_permission(env, 'p') and \
                          env.version == 'admin'):
            return False
        if hasattr(env, 'draft_form_model'):
            DraftForm = env.draft_form_model
            draft = DraftForm.get_for_item(env.db, self.stream.uid(env),
                                           item, env.user)
            if draft:
                return True
        return item.has_unpublished_changes


class DeleteFlagHandler(DeleteItemHandler):

    def delete_flag_handler(self, env, data):
        #insure_is_xhr(env)

        if env.request.method != 'POST':
            return DeleteItemHandler.__call__(self, env, data)

        self.stream.insure_has_permission(env, 'd')

        data.item.delete()

        log = self.stream.create_log_entry(env, data.item, 'delete')
        if log is not None:
            env.db.add(log)

        self.clear_tray(env, data.item)
        env.db.commit()

        stream_url = self.stream.url_for(env).qs_set(
                            data.filter_form.get_data())

        return env.json({'success': True,
                         'location': stream_url})
    __call__ = delete_flag_handler

    def is_available(self, env, item):
        if not hasattr(item, 'state'):
            return False
        return env.version == 'admin' and \
            DeleteItemHandler.is_available(self, env, item) and \
            item.state not in (item.ABSENT, item.DELETED)


class HasChangesListField(ListField):

    template='list_field_has_changes.html'

    def __init__(self, name='has_unpublished_changes', title='', **kwargs):
        kwargs.setdefault('link_to_item', False)
        kwargs.setdefault('transform', lambda x: x)
        kwargs.setdefault('classname', 'width-0')
        ListField.__init__(self, name, title, **kwargs)




class PublishStreamNoState(Stream):

    core_actions = [x for x in Stream.core_actions
                    if x.action not in ('delete', 'item')] + [
           PublishItemHandler(),
           DeleteFlagHandler(),
           PublishAction(),
           RevertAction(),
        ]

    list_base_template = 'lang_publish_stream.html'

    versions = (('admin', u'Редакторская версия'),
                ('front', u'Опубликованная версия'),)
    versions_dict = dict(versions)

    def uid(self, env, version=True):
        if version:
            return self.module_name
        return self.module_name + ':version=' + env.version

    def get_filter_form(self, env):
        cls = getattr(self.config, 'FilterForm', FilterForm)
        form = cls(env)
        form.model = self.get_model(env)
        return form

    @property
    def prefix_handler(self):
        @web.request_filter
        def set_models(env, data, nxt):
            #assert data.version in self.versions_dict.keys()
            env.models = getattr(env.models, data.version)
            env.version = data.version
            return nxt(env, data)

        version_prefix = web.prefix('/<any("%s"):version>' % \
                                     ('","'.join(self.versions_dict.keys())))
        #return version_prefix | set_models | \
        return super(PublishStreamNoState, self).prefix_handler |\
               version_prefix | set_models

    @cached_property
    def list_fields(self):
        fields = getattr(self.config, 'list_fields', {})
        cls = fields.__class__
        fields = list(fields.values())
        fields.insert(1, HasChangesListField())
        return cls(fields)

    @cached_property
    def item_template_name(self):
        return getattr(self.config, 'item_template', 'item_publish')

    def item_query(self, env):
        return env.db.query(self.get_model(env))

    def get_model(self, env):
        return getattr(env.models, self.config.Model)

    def get_edit_url(self, env, item):
        '''
        Checks if item belongs to the stream, and if it's true,
        returns an url to item edit page
        '''
        if hasattr(item, 'models'):
            model = getattr(item.models, self.config.Model, None)
            if model is not None and isinstance(item, model):
                item_ = self.item_query(env).filter_by(id=item.id).scalar()
                if item_ is not None:
                    return self.url_for(env, 'item', item=item.id)

    def commit_item_transaction(self, env, item, **kwargs):
        item._create_versions()
        if item._front_item is not None and (\
                not hasattr(item._front_item, 'state') or \
                item._front_item.state in [item.PRIVATE, item.PUBLIC]):
            # Do not mark NEW items as changed
            item.has_unpublished_changes = True
        Stream.commit_item_transaction(self, env, item, **kwargs)

    def url_for(self, env, name=None, **kwargs):
        kwargs.setdefault('version', getattr(env, 'version', self.versions[0][0]))
        return Stream.url_for(self, env, name, **kwargs)


class PublishStream(PublishStreamNoState):

    core_actions = [x for x in Stream.core_actions
                    if x.action not in ('delete', 'item')
                       and not isinstance(x, StreamListHandler)] + [
           PublishStreamListHandler(),
           PublishItemHandler(),
           DeleteFlagHandler(),
           PublishAction(),
           UnpublishAction(),
           RevertAction(),
        ]

    def item_query(self, env):
        query = PublishStreamNoState.item_query(self, env)
        if not getattr(env, 'absent_items', False): # XXX dirty hack
            Model = self.get_model(env)
            return query.filter(Model.existing)
        return query




# -*- coding: utf-8 -*-
import six
from time import time
import os
import struct
if six.PY3:
    import binascii

from webob.multidict import MultiDict
from iktomi.utils import cached_property
from ...forms import Form as BaseForm
from .fields import FieldBlock, DiffFieldSetMixIn


class Form(BaseForm):

    help_category = 'ItemForm'

    @property
    def stream(self):
        if self.env and self.env.stream:
            return self.env.stream

    @cached_property
    def id(self):
        '''Random ID for given form input'''
        # Time part is repeated in about 3 days period
        time_part = struct.pack('!d', time())[3:]
        if six.PY2:
            h = (time_part+os.urandom(1)).encode('hex')
        else:
            h = binascii.hexlify(time_part+os.urandom(1)).decode()
        return 'form' + h

    def get_help(self, fieldname):
        if self.env.stream:
            helpkey = "/".join(['streams',
                                self.env.stream.module_name,
                                self.help_category,
                                fieldname])
            return self.env.get_help(helpkey, self.help_category)


def _get_field_data(form, field):
    md = MultiDict()
    rv = field.from_python(form.python_data[field.name])
    field.set_raw_value(md, rv)
    return md


class ModelForm(Form, DiffFieldSetMixIn):

    # XXX It's a hack to support inheritance
    def __init__(self, env, initial={}, item=None, for_diff=False, **kwargs):
        #required to check unique field values
        self.item = item
        self.for_diff = for_diff
        Form.__init__(self, env, initial, **kwargs)

    def update_default(self, obj, name, value):
        setattr(obj, name, value)

    def update_pass(self, obj, name, value):
        pass

    def update_instance(self, obj):
        self._update_instance(obj, self.fields)

    def _update_instance(self, obj, fields):
        for field in fields:
            if field.writable:
                if isinstance(field, FieldBlock):
                    self._update_instance(obj, field.fields)
                else:
                    method = getattr(self, 'update__' + field.name, self.update_default)
                    method(obj, field.name, self.python_data[field.name])

    @classmethod
    def load_initial(cls, env, item, initial=None, **kwargs):
        if item.id is not None:
            initial = cls._load_initial(item, initial, cls.fields)
        return cls(env, initial, item=item, **kwargs)

    @classmethod
    def _load_initial(cls, item, initial, fields):
        initial = dict(initial or {})
        for field in fields:
            if isinstance(field, FieldBlock):
                initial.update(cls._load_initial(item, initial, field.fields))
            elif field.name:
                initial[field.name] = getattr(item, field.name)
        return initial



# -*- coding: utf-8 -*-

from sqlalchemy import Column, Integer, String, DateTime, Boolean
from iktomi.db.sqla.types import StringList
from iktomi.auth import encrypt_password, check_password as check_password_
from datetime import datetime
from iktomi.cms.models.base import register_model

__all__ = ['AdminUser']


@register_model('BaseModel')
def AdminUser(models):

    id = Column(Integer, primary_key=True)
    name = Column(String(250), nullable=False, default='')
    login = Column(String(32), nullable=False, unique=True)
    email = Column(String(200), index=True)
    password = Column(String(250), nullable=False)
    creation_time = Column(DateTime, default=datetime.now, nullable=False)
    roles = Column(StringList(250), nullable=False, default=[])
    active = Column(Boolean, default=True)

    def __unicode__(self):
        if self.id is None:
            return u'Новый редактор'
        return u'Редактор: %s (%s)' % (self.name, self.login)

    def set_password(self, password):
        self.password = encrypt_password(password)

    def check_password(self, password):
        return check_password_(password, self.password)




# -*- coding: utf-8 -*-
import six
import os
import logging
from time import time
import sqlalchemy.orm.util
if six.PY3:
    import binascii

logger = logging.getLogger('iktomi.cms.item_lock')


class ModelLockError(Exception):

    def __str__(self):
        return u'Возникли проблемы с блокировкой объекта'


class ModelLockedByOther(ModelLockError):

    def __init__(self, user, edit_session):
        ModelLockError.__init__(self, user)
        self.user = user
        self.edit_session = edit_session

    def __str__(self):
        return u'Объект заблокирован пользователем: %s (%s)' % \
                                (self.user.name, self.user.login)


class ModelLockIsLost(ModelLockError):

    def __str__(self):
        return u'Блокировка объекта утеряна'


class BaseItemLock(object):

    _lock_prefix = 'lock-'

    def __init__(self, env):
        self.env = env

    def _create_edit_session(self):
        if six.PY2:
            return os.urandom(5).encode('hex')
        else:
            return binascii.hexlify(os.urandom(5)).decode()


    @staticmethod
    def item_global_id(obj, view_in_obj=True):
        # XXX this look to the item is hacky
        if hasattr(obj, 'item_global_id') and view_in_obj:
            return obj.item_global_id()
        result = sqlalchemy.orm.util.identity_key(instance=obj)
        cls = result[0]
        ident = result[1]
        ident = '-'.join(map(str, ident))
        return '%s.%s:%s' % (cls.__module__, cls.__name__, ident)

    def _item_lock_key(self, obj):
        '''Construct key for memcache. obj should be either model object or its
        global identifier.'''
        if not isinstance(obj, six.string_types):
            obj = self.item_global_id(obj)
        return self._lock_prefix + str(obj)

    def _item_lock_value(self, edit_session):
        # edit_session - unique identifier of one open window
        # user_id - to show who locked the object
        # time - we need it since memcache doesn't guarantee it will prune a
        #        record imediately after timeout.
        return dict(edit_session=edit_session,
                    user_id=self.env.user.id,
                    time=time())

    def create(self, obj, force=False):
        '''Marks model object as editted. Returns edit session on success or
        raises exception. obj should be either model object or its global
        identifier. When force is True the current lock is ignored.'''
        raise NotImplementedError()

    def update(self, obj, edit_session):
        '''Updates model object lock as being active. Raises exception on
        error. obj should be either model object or its global identifier.'''
        raise NotImplementedError()

    def update_or_create(self, obj, edit_session):
        '''Updates item lock if edit_session is empty, otherwise creates a new
        lock. Returns edit_session if success or raises exception if failed'''
        if edit_session:
            self.update(obj, edit_session)
            return edit_session
        else:
            return self.create(obj)

    def remove(self, obj, edit_session):
        '''Removes lock for model object. obj should be either model object or
        its global identifier.'''
        raise NotImplementedError()

    def check(self, obj):
        raise NotImplementedError()


class ItemLockData(object):

    def __init__(self, env, stream, item, filter_form,
                 edit_session, owner_session, message):
        self.env, self.filter_form, self.stream, self.item = \
                env, filter_form, stream, item
        self.edit_session, self.owner_session, self.message = \
                edit_session, owner_session, message

    def render(self):
        env, stream = self.env, self.stream
        back_url = stream.lock_back_url(env, self.item, self.filter_form)
        global_id = BaseItemLock.item_global_id(self.item)
        tdata = dict(self.__dict__,
                     global_id=global_id,
                     lock_timeout=env.cfg.MODEL_LOCK_RENEW,
                     back_title = stream.lock_back_title,
                     back_help = stream.lock_back_help,
                     back_url = back_url)
        return env.render_to_string("macros/item_lock", tdata)

    @classmethod
    def for_item(cls, env, stream, item, filter_form=None):
        request = env.request
        edit_session = request.POST.get('edit_session',
                                        request.GET.get('edit_session', ''))
        owner_session = message = ''
        if item is not None and item.id is not None:
            try:
                edit_session = env.item_lock.update_or_create(
                    item, edit_session)
            except ModelLockedByOther as e:
                message = six.u(e)
                owner_session = e.edit_session
            except ModelLockError as e:
                message = six.u(e)
        return cls(env, stream, item, filter_form,
                   edit_session, owner_session, message)




# -*- coding: utf-8 -*-
import six

from iktomi import web
from iktomi.utils.deprecation import deprecated


@web.request_filter
@deprecated('Flash messages are rendered in templates and JSON now')
def flash_message_handler(env, data, nxt):
    return nxt(env, data)


def flash(env, message, category=None, unique=True):
    # XXX this will not work on errors
    if not hasattr(env, '_flash'):
        env._flash = []
    value = (message, category)
    if not unique or value not in env._flash:
        env._flash.append(value)



# -*- coding: utf-8 -*-

from iktomi import web
from iktomi.utils import cached_property
from iktomi.forms import Form, Field, FieldSet, FieldList
from iktomi.cms.forms import convs

from .stream_actions import StreamAction
from .item_lock import ModelLockedByOther, ModelLockError
from .flashmessages import flash
from webob.exc import HTTPForbidden


class ListItemModelChoice(convs.ModelChoice):

    @property
    def model(self):
        return self.env.stream.get_model(self.env)


class ListItemForm(Form):

    template = 'stream_sortables.html'

    def __init__(self, *args, **kwargs):
        kwargs['initial'] = self.initial_for_items(kwargs.pop('items', []))
        Form.__init__(self, *args, **kwargs)

    fields = [
            FieldList(
                'items',
                widget=FieldList.widget(
                    template='list_item_form'),
                field=FieldSet(
                    'item',
                    fields=[
                        Field('item',
                              conv=ListItemModelChoice(required=False),
                              label=u'item'),
                    ]
                ),
            ),
        ]

    def initial_for_item(self, item):
        return {'item': item}

    def initial_for_items(self, items):
        return {'items': [self.initial_for_item(item)
                          for item in items]}


class ListEditAction(StreamAction):

    action='list_edit'
    item_lock=True
    display = False
    for_item = False
    title=u'Редактировать'
    use_with_filters = False

    def save_allowed(self, env):
        return self.stream.has_permission(env, 'w')

    @cached_property
    def ListItemForm(self):
        return getattr(self.stream.config, 'ListItemForm', ListItemForm)

    def __call__(self, env, data):
        raise NotImplementedError

    @property
    def app(self):
        return web.method('POST') | web.match('/edit', 'list_edit') | self





# -*- coding: utf-8 -*-
from iktomi import web
from iktomi.cms.stream import Stream


class I18nStream(Stream):

    langs = [('ru', u'Русский'),
             ('en', u'Английский')]
    langs_dict = dict(langs)
    items_are_i18n = True
    list_base_template = 'lang_publish_stream.html' # XXX ?

    def uid(self, env, version=False):
        # Attention! Be careful!
        # Do not change format of uid unless you are sure it will not 
        # brake tray views, where stream_name and language are parsed out
        # from the uid
        return self.module_name + ':lang=' + env.lang

    @property
    def prefix_handler(self):
        @web.request_filter
        def set_models(env, data, nxt):
            env.models = getattr(env.models, data.lang)
            env.lang = data.lang
            return nxt(env, data)

        lang_prefix = web.prefix('/<any("%s"):lang>' % \
                                     ('","'.join(self.langs_dict.keys())))
        return Stream.prefix_handler.fget(self) | lang_prefix | set_models

    def url_for(self, env, name=None, **kwargs):
        kwargs.setdefault('lang', getattr(env, 'lang', None) or self.langs[0][0])
        return Stream.url_for(self, env, name, **kwargs)

    def get_model(self, env):
        return getattr(getattr(env.models, env.lang),
                       self.config.Model)



# -*- coding: utf-8 -*-
from .stream import PublishStream
from .i18n_stream import I18nPublishStream
from iktomi.web import WebHandler
from iktomi.utils.storage import VersionedStorage


class PublishQueue(WebHandler):

    def __init__(self, streams):
        self.streams = streams

    def publish_queue(self, env, data):
        #insure_is_xhr(env)

        changed = []
        for stream in self.streams.values():
            if isinstance(stream, I18nPublishStream):
                for lang, lang_name in stream.langs:
                    subenv = VersionedStorage()
                    subenv._parent_storage = env
                    subenv.models = getattr(env.models.admin, lang)
                    subenv.version = 'admin'
                    subenv.lang = lang

                    items = stream.item_query(subenv)\
                                  .filter_by(has_unpublished_changes=True)
                    if hasattr(stream.config.Model, 'state'):
                        items = items.filter_by(state=stream.config.Model.PUBLIC)
                    items = items.all()
                    changed += [(stream, subenv, item) for item in items]
            elif isinstance(stream, PublishStream):
                subenv = VersionedStorage()
                subenv._parent_storage = env
                subenv.models = env.models.admin
                subenv.version = 'admin'
                items = stream.item_query(subenv)\
                              .filter_by(has_unpublished_changes=True)
                if hasattr(stream.config.Model, 'state'):
                    items = items.filter_by(state=stream.config.Model.PUBLIC)
                items = items.all()
                changed += [(stream, subenv, item) for item in items]
        changed.sort(key=lambda x: x[2].updated_dt)
        return env.render_to_response('publish_queue', dict(
            changed = changed,
            menu = env.current_location,
            title = u'Очередь публикации',
        ))
    __call__ = publish_queue



# -*- coding: utf-8 -*-
import six

from webob.exc import HTTPNotFound

from iktomi import web
from iktomi.cms.stream_actions import GetAction
from iktomi.cms.loner import Loner


class PreviewHandler(GetAction):

    item_lock = True
    allowed_for_new = False
    cls = 'preview'
    action = 'preview'
    title = u'Предпросмотр'

    @property
    def PrepareItemHandler(self):
        return self.stream.edit_action.PrepareItemHandler

    @property
    def app(self):
        if isinstance(self.stream, Loner):
            prefix = web.prefix('/'+self.action, name=self.action)
        else:
            prefix = web.prefix('/<idconv:item>/'+self.action, name=self.action,
                                convs={'idconv': self.stream.id_converter})
        return prefix | self.PrepareItemHandler(self) | self

    def item_url(self, env, data, item):
        raise NotImplementedError

    def external_url(self, env, data, item):
        raise NotImplementedError

    def is_available(self, env, item=None):
        return GetAction.is_available(self, env, item) and \
                getattr(env, 'version', None) != 'front'

    def preview(self, env, data):
        #insure_is_xhr(env)
        if not self.is_available(env, data.item):
            raise HTTPNotFound

        item = data.item

        tdata = dict(filter_form=data.filter_form,
                     roles=env.user.roles,
                     item=item,
                     item_url=self.item_url(env, data, item),
                     item_lock = data.item_lock,
                     external_url=self.external_url(env, data, item),
                     stream=self.stream,
                     stream_url=self.stream.url_for(env),
                     stream_title=self.stream.config.title,
                     title=six.u(item),
                     menu=self.stream.module_name,
                     actions=[x for x in self.stream.actions 
                              if x.for_item and x.is_visible(env, item)
                                and x.action not in ('preview', 'delete')])
        return env.render_to_response('preview', tdata)

    __call__ = preview





# -*- coding: utf-8 -*-
from iktomi import web
from webob.exc import HTTPForbidden


class StreamAction(web.WebHandler):

    item_lock = True
    accepts_item_form = True
    for_item = True
    action = title = cls = js_block = None
    allowed_for_new = False
    display = True
    hint = None
    xhr = True

    def is_available(self, env, item=None):
        """ Rewrite this method to add condition when action is avalible
        """
        if item is None or item.id is None:
            return self.allowed_for_new
        return True

    def is_visible(self, env, item=None):
        """ Tells if this action button is visible.
        """
        return self.is_available(env, item) and self.display

    def insure_is_available(self, env, item=None):
        """ Shortcut to check is_avalible rule on handle
        """
        if not self.is_available(env, item):
            raise HTTPForbidden

    def __init__(self, stream=None, **kw):
        self.init_kwargs = kw
        self.stream = stream
        self.action = kw.get('action', self.action)
        self.title = kw.get('title', self.title)
        self.cls = kw.get('cls', self.cls)
        self.allowed_for_new = kw.get('allowed_for_new', self.allowed_for_new)
        self.display = kw.get('display', self.display)
        self.hint = kw.get('hint', self.hint)
        if stream is not None and \
                getattr(stream.config, 'item_lock', None) == False:
            self.item_lock = False

    def bind(self, stream):
        return self.__class__(stream=stream, **self.init_kwargs)

    def url(self, env, item=None):
        kwargs = {}
        if item:
            kwargs['item'] = item.id
        return self.stream.url_for(env, self.action, **kwargs)

    def help_message(self, env):
        action_name = self.action or self.__class__.__name__
        helpkey = "/".join(['streams',
                            env.stream.module_name,
                            'Action',
                            action_name])
        return env.get_help(helpkey, 'Action')



class PostAction(StreamAction):

    mode = 'post'


class GetAction(StreamAction):

    mode = 'get'


class CustomAction(StreamAction):

    mode = 'custom'


class AfterPostAction(StreamAction):

    mode = 'after-post'





# -*- coding: utf-8 -*-
from sqlalchemy import Column, ForeignKey, Integer, PickleType, DateTime, \
        Index, String
from sqlalchemy.dialects.mysql import MEDIUMBLOB
from sqlalchemy.orm import relationship
from datetime import datetime
from .base import register_model

__all__ = ['DraftForm', 'DraftFormAdminUser']


# XXX MySQL-specific type. How to resulve this?
class MediumPickleType(PickleType):

    impl = MEDIUMBLOB


@register_model('BaseModel')
def DraftFormAdminUser(models):

    admin_id  = Column(ForeignKey(models.AdminUser.id,
                                  ondelete="CASCADE"),
                       primary_key=True)
    draft_id  = Column(Integer, ForeignKey(models.DraftForm.id,
                                           ondelete="CASCADE"),
                       primary_key=True)


@register_model('BaseModel')
def DraftForm(models):

    id = Column(Integer, primary_key=True)
    stream_name = Column(String(50), nullable=False, default='')
    # object id can be string, so we use string here
    object_id = Column(String(50), nullable=True)
    data = Column(MediumPickleType, default=[])
    creation_time = Column(DateTime, default=datetime.now, nullable=False)
    update_time = Column(DateTime, default=datetime.now, nullable=False)

    users = relationship(models.AdminUser,
                          secondary=models.DraftFormAdminUser.__table__)

    @classmethod
    def get_for_item(cls, db, stream_name, item, user):
        '''Get DraftForm instance related to given item and stream.
           If the item is None, then DraftForm instances are filtered
           also by users
        '''
        query = db.query(cls).filter_by(stream_name=stream_name)
        if item is None or item.id is None:
            return query.filter(cls.users.contains(user) &
                                (cls.object_id == None)).first()
        return query.filter(cls.object_id == str(item.id)).first()

    Index('draft_index', stream_name, object_id)



from .base import model_factories

from .autosave import *
from iktomi.cms.auth.models import *
from iktomi.cms.tray.models import *
from iktomi.cms.editor_notes.models import *
from iktomi.cms.edit_log.models import *

__all__ = ['model_factories']




# -*- coding: utf-8 -*-
import six
from iktomi.forms.widgets import *
from iktomi.forms import widgets

from datetime import datetime
from iktomi.cms.forms import convs
from iktomi.utils import cached_property
from iktomi.cms.edit_log.models import make_diff
from lxml.html.diff import htmldiff
import json

class _LazyHtmlDiff(object):
    def __init__(self, value1, value2):
        self.value1, self.value2 = value1, value2

    @cached_property
    def value(self):
        return htmldiff(self.value1, self.value2)


class WysiHtml5(Widget):

    template = 'widgets/wysihtml5'

    button_blocks = [
        ('inline', ['bold', 'italic', 'underline']),
        ('block', ['headings', 'sup', 'sub', 'blockquote']),
        ('lists', ['insertunorderedlist', 'insertorderedlist',
                   'outdent', 'indent']),
        ('advanced', ['createLink', 'insertImage', 'table', 'extrachars']),
        ('history', ['undo', 'redo']),
        ('html', ['html']),
    ]

    BUTTON_TAGS = (
        ('bold', 'b'),
        ('italic', 'i'),
        ('underline', 'u'),
        ('sup', 'sup'),
        ('sub', 'sub'),
        ('blockquote', 'blockquote'),
        ('insertunorderedlist', ('ul', 'li')),
        ('insertorderedlist', ('ol', 'li')),
        ('outdent', ('ul', 'li')),
        ('indent', ('ul', 'li')),
        ('outdent', ('ol', 'li')),
        ('indent', ('ol', 'li')),

        ('createLink', 'a'),
        ('insertImage', 'img'),

        ('table', ('table', 'td', 'tr')),

        ('headings', 'h1'),
        ('headings', 'h2'),
        ('headings', 'h3'),
        ('headings', 'h4'),
        ('h1', 'h1'),
        ('h2', 'h2'),
        ('h3', 'h3'),
        ('h4', 'h4'),
        ('aside', 'aside'),
    )

    stylesheets = ("/static/css/wysihtml5-content.css",)

    @cached_property
    def js_config(self):
        return json.dumps({'parserRules': self.parser_rules,
                           'cleanerConfig': self.cleaner_config,
                           'stylesheets': self.stylesheets})

    @cached_property
    def parser_rules(self):
        rules = {}
        attrs = {}
        for attr in self.html_conv.allowed_attributes:
            attrs[attr] = 'none'
        for tag in self.html_conv.allowed_elements:
            rules[tag] = {'check_attributes': attrs}

        if 'b' in rules and not 'strong' in rules:
            rules['strong']  = {'rename_tag': 'b'}
        if 'strong' in rules and not 'b' in rules:
            rules['b']  = {'rename_tag': 'strong'}
        if 'em' in rules and not 'i' in rules:
            rules['i']  = {'rename_tag': 'em'}
        if 'i' in rules and not 'em' in rules:
            rules['em']  = {'rename_tag': 'i'}
        return {'tags': rules}

    @cached_property
    def cleaner_config(self):
        conv = self.field.conv
        config = {}
        if hasattr(conv, 'tags_to_wrap'):
            config['tagsToWrap'] = [x.upper() for x in conv.tags_to_wrap]
        if hasattr(conv, 'drop_empty_tags'):
            config['dropEmptyTags'] = [x.upper() for x in conv.drop_empty_tags]
        if hasattr(conv, 'wrap_inline_tags'):
            config['wrapInlineTags'] = conv.wrap_inline_tags
        if hasattr(conv, 'split_paragraphs_by_br'):
            # XXX does not take an effect in the converter
            config['splitParagraphsByBr'] = conv.split_paragraphs_by_br
        return config

    @cached_property
    def html_conv(self):
        html_conv = self.field.conv
        if isinstance(html_conv, convs.ListOf):
            html_conv = html_conv.conv
        assert isinstance(html_conv, convs.Html)
        return html_conv

    @cached_property
    def allowed_elements(self):
        return set(self.html_conv.allowed_elements)

    @cached_property
    def real_buttons(self):
        buttons = set()
        tag_buttons = set()
        for button, tags in self.BUTTON_TAGS:
            tag_buttons.add(button)
            tags = [tags] if isinstance(tags, six.string_types) else tags
            if not (set(tags) - self.allowed_elements):
                buttons.add(button)
        all_btns = set(sum(dict(self.button_blocks).values(), []))
        allowed = (set(all_btns) - tag_buttons) | buttons
        buttons = [[btn for btn in btns if btn in allowed]
                   for name, btns in self.button_blocks]
        return list(filter(None, buttons))

    def has_button(self, button):
        return button in self.real_buttons

    def remove_buttons(self, remove_buttons):
        buttons = [
                (name, [btn for btn in btns 
                        if btn not in remove_buttons])
                for name, btns in self.button_blocks]
        buttons = list(filter(lambda x: x[1], buttons))
        return self(button_blocks=buttons)

    def add_buttons(self, buttons):
        button_blocks = [
                (name, [btn for btn in btns if btn not in buttons])
                for name, btns in self.button_blocks]
        buttons_dict = dict(button_blocks)
        for name, btns in buttons:
            if name in buttons_dict:
                buttons_dict[name] += btns
            else:
                button_blocks.append((name, btns))
        return self(button_blocks=button_blocks)

    def prepare_data(self):
        return dict(Widget.prepare_data(self),
                    js_config=self.js_config)

    def render_as_diff(self, diff, marker):
        '''
        Renders widget to template
        '''
        data = self.prepare_data()
        data['value'] = diff
        rev_marker = 'ins' if marker == 'del' else 'del'
        parser_rules = dict(self.parser_rules)
        parser_rules['tags'] = dict(self.parser_rules['tags'],
                            **{rev_marker: {'remove': 1},
                               marker: {}})
        data['js_config'] = json.dumps({'parserRules': parser_rules,
                                        'cleanerConfig': self.cleaner_config,
                                        'stylesheets': self.stylesheets})
        if self.field.readable:
            return self.env.template.render(self.template, **data)
        return ''

    @staticmethod
    def get_diff(field1, field2):
        data1 = field1.from_python(field1.clean_value)
        data2 = field2.from_python(field2.clean_value)

        if field1 is None or field2 is None:
            if data1 != data2:
                return make_diff(field1, field2,
                                 changed=True)
        elif data1 != data2:
            diff = _LazyHtmlDiff(data1, data2)
            before=lambda: field1.widget.render_as_diff(diff.value, 'del')
            after=lambda: field2.widget.render_as_diff(diff.value, 'ins')
            return dict(label=field1.label or field1.name,
                        name=field1.name,
                        before=before,
                        after=after,
                        changed=True)



class PopupStreamSelect(Select):

    _obsolete = Select._obsolete | frozenset(['reorderable'])

    template = 'widgets/popup_stream_select'
    open_btn_text = u'Выбрать'
    create_btn_text = u'Создать'
    allow_create = False
    allow_select = True
    allow_delete = True
    sortable = True
    select_all_button = True
    unshift = False
    default_filters = {}
    rel = None # set rel="popup" to open and edit the item in popup

    @cached_property
    def default_create_filters(self):
        return self.default_filters

    @cached_property
    def stream(self):
        return self.env.streams[self.stream_name]

    def render_row_template(self, **data):
        return self.env.render_to_string(self.stream.row_template_name,
                                         dict(self.stream.template_data, **data))

    def item_row(self, item, row_cls=''):
        url = self.stream.url_for(self.env, 'item', item=item.id)
        read_allowed=self.stream.has_permission(self.env, 'r')
        return self.render_row_template(stream=self.stream,
                                        item=item,
                                        list_fields=self.list_fields,
                                        read_allowed=read_allowed,
                                        url=url, row_cls=row_cls)

    @cached_property
    def list_fields(self):
        return self.stream.list_fields

    @cached_property
    def create_url(self):
        return self.stream.url_for(self.env, 'item', item=None)\
                          .qs_set(self.default_create_filters)

    def js_config(self):
        data = {
            'url': self.stream.url_for(self.env).qs_set(self.default_filters),
            'title': self.stream.title,
            'container': self.id,
            'input_name': self.input_name,
            'allow_delete': self.allow_delete,
            'sortable': self.sortable,
            'unshift': self.unshift,
        }
        if self.rel is not None:
            data['rel'] = self.rel
        if self.allow_create:
            data['create_url'] = self.create_url
        return json.dumps(data) # XXX escape_js

    @cached_property
    def show_create_button(self):
        return self.allow_create and self.stream.has_permission(self.env, 'c')

    def get_options(self, value):
        choice_conv = self.field.conv
        if isinstance(choice_conv, convs.ListOf):
            choice_conv = choice_conv.conv
        assert isinstance(choice_conv, convs.EnumChoice)

        values = value if self.multiple else [value]
        values = list(filter(None, map(choice_conv.to_python, values)))
        return values


class PopupFilteredSelect(Select):

    template = 'widgets/popup_filtered_select'
    open_btn_text = u'Выбрать'
    disable_unpublished = False

    def get_options(self, value):
        options = []
        # XXX ugly
        choice_conv = self.field.conv
        if isinstance(choice_conv, convs.ListOf):
            choice_conv = choice_conv.conv
        assert isinstance(choice_conv, convs.EnumChoice)

        values = value if self.multiple else [value]
        values = map(six.u, values)
        for choice, label in choice_conv.options():
            choice = six.u(choice)
            options.append(dict(value=choice,
                                title=label,
                                selected=(choice in values)))
        return options

    def js_config(self):
        return json.dumps({
            'multiple': self.multiple,
            'required': self.field.conv.required,
            'disable_unpublished': self.disable_unpublished,
            'readonly': not self.field.writable,
        })


class TabSelect(Select):

    template = 'widgets/tab_select'
    inject_to = 'list_tabs'

    def js_config(self):
        return json.dumps(dict(
            getattr(self, 'js_conf', {}),
            inject_to=self.inject_to
        ))


class LabelSelect(Select):

    template = 'widgets/label_select'
    hiddens = [] # hide these options if they are not set


class AjaxFileInput(FileInput):

    template = 'widgets/ajax_fileinput'
    upload_url = None

    def js_config(self):
        conf = {'input': self.id,
                'error': self.field.error,
                'upload_url': self.upload_url,
                'value': self.field.clean_value}

        # XXX copy-paste from old code. Should be replaced with better alternative after new uploading api stabilization.
        if self.field.clean_value is not None:
            conf['is_value_persistent'] = self.field.clean_value.mode == 'existing'
        return json.dumps(conf)


class Calendar(TextInput):

    template = 'widgets/calendar'
    classname = 'calendar'
    js_block = 'calendar'

    @cached_property
    def size(self):
        return len(self.field.from_python(datetime(1999, 12, 31)))+1


class FieldBlockWidget(widgets.FieldBlockWidget):
    """
    If FieldBlock needs to be collapsed by default,
    pass `closed=True` to the widget::

        FieldBlock('name',
                   fields=[..],
                   widget=FieldBlock.widget(closed=True))
    """

    classname = 'collapsable'
    template = 'widgets/collapsable_block'
    renders_hint = True
    closed = False


class FieldSetWidget(widgets.FieldSetWidget):
    js_block = None



# -*- coding: utf-8 -*-
import os
import logging
from jinja2 import Markup

logger = logging.getLogger(__name__)


class HelpLoader(object):
    '''
    Helploader is needed to
      1. Load help messages from directory with help templates.
           `helploader = HelpLoader("path/to/help/messages").`
      2. Retrieving help messages with get_help method.
         It is supposed to add get_help message to your Environment:
           `
           class Environment(BasicEnvironment):
               get_help = helploader.get_help
           `
         Calling get_help requires 2 parameters: key and category.
         Key is the name of the element on screen.
         Category means part of the interface, where element is located.
         For example, when get_help is called this way:
               `get_help('streams/docs/ItemForm/title', 'ItemForm')`
         helploader tries to get help-messages from
         `streams/docs/ItemForm/title.html` if if it exists, and, if not,
         from `ItemForm.default/title.html`, where `ItemForm.default` is
         default directory for messages with `ItemForm` category.
         Structure of help directory:
           `
           /..
           /Category_name.default/*.html - all default mesages for Category_name
           /streams/stream_module_name/Stream_category/*.html - specific messages for stream_module_name stream
           `
    '''
    help_messages = {}
    default_help_messages = {}

    def __init__(self, template_root):
        self.template_root = template_root
        self.load_help()

    def load_default_help(self, dirname):
        dirpath = os.path.join(self.template_root, dirname)
        messages = {}
        for filename in os.listdir(dirpath):
            if filename.startswith("."):
                continue
            filepath = os.path.join(self.template_root, dirname, filename)
            with open(filepath) as f:
                helpkey, _ = os.path.splitext(filename)
                messages[helpkey] = f.read().decode('utf8')
        category, _ = os.path.splitext(dirname)
        self.default_help_messages[category] = messages

    def load_help(self):
        if not os.path.exists(self.template_root):
            logger.warning("Help-dir %s not exists!", self.template_root)
            return {}

        for root, dirs, files in os.walk(self.template_root):
            if root == self.template_root:
                for dirname in dirs:
                    if dirname.endswith("default"):
                        self.load_default_help(dirname)
            if files:
                for filename in files:
                    if filename.startswith("."):
                        continue
                    template_name = os.path.join(root[len(self.template_root):], filename)
                    filepath = os.path.join(root, filename)
                    helpkey, _ = os.path.splitext(template_name)
                    with open(filepath) as f:
                        self.help_messages[helpkey.lstrip("/")] = f.read().decode('utf8')

    def get_help(self, key, category='Global'):
        html = self.help_messages.get(key, '')
        if html is '':
            key = key.split('/')[-1]
            html = self.default_help_messages.get(category, {}).get(key, '')
        return Markup(html)



# -*- coding: utf-8 -*-

import re
import json
import logging, binascii
from jinja2 import Markup
from functools import partial
from os import path
from iktomi import web
from iktomi.web import WebHandler
from io import open
import six

from webob.exc import HTTPNotModified


logger = logging.getLogger(__name__)

def is_absolute(url):
    return url.startswith('/') or url.startswith('http://') \
                               or url.startswith('https://')

class StaticPacker(WebHandler):

    def response_with_etag(self, request, content, headers=None,
                           check=None, mimetype='text/html'):
        headers = headers or {}
        check = check or content
        etag_request = request.headers.get('If-None-Match')
        # XXX may be another algorythm?
        if isinstance(check, six.text_type):
            check = check.encode('utf-8')
        else:
            check = check.decode()
        etag_response = str(binascii.crc32(check))
        if etag_request == etag_response:
            raise HTTPNotModified()

        headers['ETag'] = etag_response
        response = web.Response(content)
        response.headers['Content-Type'] = mimetype
        for name, value in headers.items():
            response.headers[name] = value
        return response

    def import_css(self, base_url, base_path, match):
        url = match.group(1)
        media = match.group(2)
        if is_absolute(url):
            return '@import url("%s") %s;' % (url, media)
        return self.get_css_contents(base_url, base_path, path.join(base_path, url))

    def fix_url(self, base_url, match):
        url = match.group(1).strip('"\'')
        if is_absolute(url):
            return 'url(%s)' % url
        url = base_url + url
        return 'url(%s)' % url

    def get_css_contents(self, base_url, base_path, filename):
        with open(filename, 'rb') as f:
            data = f.read().decode('utf-8').strip()
            base_path = path.dirname(filename)
            data = re.sub(r'@import (?:url\()?"?\'?([^"\')]+)\'?"?\)?\s*(\w*)\s*;',
                          partial(self.import_css, base_url, base_path),
                          data)

            data = re.sub(r'url\(([^\)]+)\)',
                          partial(self.fix_url, base_url),
                          data)

            return data

    def css_packer(self, env, data):
        if getattr(env.cfg, 'CACHE_PACKED', False) \
                                and getattr(self, '_cached_css', None):
            return self.response_with_etag(env.request, self._cached_css,
                                                {'From-Cache': 'TRUE'},
                                                mimetype='text/css')
        content = []
        for manifest in env.cfg.MANIFESTS.values():
            manifest_filename = path.join(manifest['path'], manifest['css'])
            root = path.dirname(manifest_filename)
            base_url = manifest['url'] + path.dirname(manifest['css']) + '/'
            with open(manifest_filename) as f_manifest:
                files = [x.split('#')[0].strip()
                         for x in f_manifest.read().splitlines()]
                files = filter(None, files)
                files = [x + '.css' for x in files]
            for name in files:
                data = self.get_css_contents(base_url, root, path.join(root, name))
                content.append('/* '+ name +' */\n\n'+data)

        content = u'\n\n'.join(content)
        self._cached_css = content
        return self.response_with_etag(env.request, content, mimetype='text/css')


    def js_packer(self, env, data):
        if getattr(env.cfg, 'CACHE_PACKED', False)  \
                                and getattr(self, '_cached_js', None):
            return self.response_with_etag(env.request, self._cached_js,
                                           {'From-Cache': 'TRUE'},
                                           mimetype='text/javascript')

        content = []
        for manifest in env.cfg.MANIFESTS.values():
            manifest_filename = path.join(manifest['path'], manifest['js'])
            root = path.dirname(manifest_filename)
            with open(manifest_filename) as f_manifest:
                files = [x.split('#')[0].strip()
                         for x in f_manifest.read().splitlines()]
                files = filter(None, files)
                files = [x +'.js' for x in files]
            for name in files:
                filename = path.join(root, name)
                with open(filename, 'rb') as f:
                    data = f.read().decode('utf-8').strip()
                    if not data.endswith(';'):
                        data += ';'
                    content.append(u'/* %s */\n%s' % (json.dumps(name), data))

        content.insert(0, "// Scripts: " + ",".join(files))
        content = u'\n\n'.join(content)
        self._cached_js = content
        return self.response_with_etag(env.request, content, mimetype='text/javascript')

    def _read_manifests(self, env, tp):
        result = []
        for manifest in env.cfg.MANIFESTS.values():
            manifest_filename = path.join(manifest['path'], manifest[tp])

            out = []
            with open(manifest_filename, 'r') as manifest_file:
                for line in manifest_file:
                    line = line.split('#')[0].strip()
                    if line:
                        out.append(line)

            base_url = manifest['url'] + path.dirname(manifest[tp])
            out = ['%s/%s.%s' % (base_url, name, tp) for name in out]
            result += out
        return result

    def js_tag(self, env):
        if getattr(env.cfg, 'RAW_JS', False):

            out = self._read_manifests(env, 'js')
            out = ['<script type="text/javascript" src="%s"></script>' % url
                   for url in out]
            return Markup('\n'.join(out))
        else:
            url = env.url_for('js_packer')
            return Markup('<script type="text/javascript" src="%s"></script>' % url)

    def css_tag(self, env, media='screen, projection', doctype="xhtml"):
        close = "/" if doctype.lower() == 'xhtml' else ""
        if getattr(env.cfg, 'RAW_CSS', False):

            out = self._read_manifests(env, 'css')
            out = ['<link rel="stylesheet" type="text/css" '
                   'media="%s" href="%s" %s>' % (media, url, close)
                   for url in out]
            return Markup('\n'.join(out))
        else:
            url = env.url_for('css_packer')
            return Markup('<link rel="stylesheet" type="text/css" '\
                          'media="%s" href="%s" %s>' % (media, url, close))



# -*- coding: utf-8 -*-

from distutils.core import setup

setup(
    name='iktomi.cms',
    version='0.2.11',
    packages=['iktomi',
              'iktomi.cms',
                    'iktomi.cms.auth',
                    'iktomi.cms.edit_log',
                    'iktomi.cms.editor_notes',
                    'iktomi.cms.forms',
                    'iktomi.cms.item_lock',
                    'iktomi.cms.models',
                    'iktomi.cms.publishing',
                    'iktomi.cms.tray'],

    package_dir={
        'iktomi.cms': 'iktomi/cms'
    },
    package_data={
        'iktomi.cms': ['templates/*.html',
                       'templates/*/*.html',
                       'templates/*/*/*.html',
                       'static/js/*.js',
                       'static/js/*.json',
                       'static/js/*/*.js',
                       'static/js/Manifest',
                       'static/css/*.css',
                       'static/css/*/*.css',
                       'static/css/Manifest',
                       'static/images/*.*',
                       'static/images/*/*.*',
                       ],
    },
    requires=[
        'webob (>1.1b1)',
        'iktomi (>0.4.2)',
    ],
    author='Denis Otkidach',
    author_email='denis.otkidach@gmail.com',
    maintainer='Evgeny Mayorov',
    maintainer_email='motormen@gmail.com',
    description='CMS components for iktomi.',
    #long_description=open('README').read(),
    url='http://github.com/SmartTeleMax/iktomi-cms/',
    license='MIT',
    #keywords='',
)



# -*- coding: utf-8 -*-
from __future__ import absolute_import

import json
from iktomi.cms.item_lock.base import BaseItemLock, ModelLockError, ModelLockedByOther, \
        ModelLockIsLost, logger
from redis import WatchError

def loads(value):
    if value:
        return json.loads(value.decode())


class RedisItemLock(BaseItemLock):

    def create(self, obj, force=False):
        '''Marks model object as editted. Returns edit session on success or
        raises exception. obj should be either model object or its global
        identifier. When force is True the current lock is ignored.'''
        cfg = self.env.cfg
        redis = self.env.redis

        edit_session = self._create_edit_session()
        key = self._item_lock_key(obj)
        value = self._item_lock_value(edit_session)
        if force:
            redis.set(key, json.dumps(value))
            redis.expire(key, cfg.MODEL_LOCK_TIMEOUT)
            return edit_session

        with redis.pipeline() as pipe:
            for i in range(2):
                try:
                    # put a WATCH on the key that holds our sequence value
                    pipe.watch(key)
                    # after WATCHing, the pipeline is put into immediate execution
                    # mode until we tell it to start buffering commands again.
                    # this allows us to get the current value of our sequence
                    old_value = loads(pipe.get(key))

                    if old_value is None:
                        # now we can put the pipeline back into buffered mode with MULTI
                        pipe.multi()
                        pipe.set(key, json.dumps(value))
                        pipe.expire(key, cfg.MODEL_LOCK_TIMEOUT)
                        # and finally, execute the pipeline (the set command)
                        pipe.execute()
                        # if a WatchError wasn't raised during execution, everything
                        # we just did happened atomically.
                        return edit_session
                    else:
                        lock_user = self.env.db.query(self.env.auth_model)\
                                                    .get(old_value['user_id'])
                        raise ModelLockedByOther(lock_user, old_value['edit_session'])
                except WatchError:
                    continue
        logger.error('Failed to lock model object. '\
                     'Problem with redis?')
        raise ModelLockError()


    def update(self, obj, edit_session):
        '''Updates model object lock as being active. Raises exception on
        error. obj should be either model object or its global identifier.'''
        cfg = self.env.cfg
        redis = self.env.redis
        key = self._item_lock_key(obj)

        with redis.pipeline() as pipe:
            for i in range(2):
                try:
                    # put a WATCH on the key that holds our sequence value
                    pipe.watch(key)
                    old_value = loads(pipe.get(key))

                    if old_value is None:
                        raise ModelLockIsLost()
                    elif old_value['edit_session']==edit_session:
                        # now we can put the pipeline back into buffered mode with MULTI
                        pipe.multi()
                        pipe.expire(key, cfg.MODEL_LOCK_TIMEOUT)
                        pipe.execute()
                        return edit_session
                    else:
                        lock_user = self.env.db.query(self.env.auth_model)\
                                                    .get(old_value['user_id'])
                        raise ModelLockedByOther(lock_user, old_value['edit_session'])
                except WatchError:
                    continue
        logger.error('Failed to update model object lock. '\
                     'Problem with redis?')
        raise ModelLockError()

    def remove(self, obj, edit_session):
        '''Removes lock for model object. obj should be either model object or
        its global identifier.'''
        redis = self.env.redis
        key = self._item_lock_key(obj)

        with redis.pipeline() as pipe:
            try:
                # put a WATCH on the key that holds our sequence value
                pipe.watch(key)
                old_value = loads(pipe.get(key))

                if old_value is not None and old_value['edit_session']==edit_session:
                    # now we can put the pipeline back into buffered mode with MULTI
                    pipe.multi()
                    pipe.delete(key)
                    pipe.execute()
                return
            except WatchError:
                logger.info('Collision on lock removal')

    def check(self, obj):
        redis = self.env.redis
        key = self._item_lock_key(obj)
        value = loads(redis.get(key))
        return value or None
